import setuptools.command.build


# pylint: disable-next=invalid-name
class build_extra(setuptools.command.build.build):
    """Adds the extra commands to the build target. This class should be
    used with setuptools."""

    def __init__(self, dist):
        super().__init__(dist)

        self.user_options.extend(
            [
                ("i18n", None, "use the localisation"),
                ("icons", None, "use icons"),
                ("kdeui", None, "use kdeui"),
                ("help", None, "use help system"),
            ]
        )

    def initialize_options(self):
        super().initialize_options()
        self.i18n = False
        self.icons = False
        self.help = False
        self.kdeui = False

    def finalize_options(self):
        def has_help(unused_command):
            return self.help == "True" or (
                "build_help" in self.distribution.cmdclass and self.help != "False"
            )

        def has_icons(unused_command):
            return self.icons == "True" or (
                "build_icons" in self.distribution.cmdclass and self.help != "False"
            )

        def has_i18n(unused_command):
            return self.i18n == "True" or (
                "build_i18n" in self.distribution.cmdclass and self.i18n != "False"
            )

        def has_kdeui(unused_command):
            return self.kdeui == "True" or (
                "build_kdeui" in self.distribution.cmdclass and self.kdeui != "False"
            )

        super().finalize_options()
        self.sub_commands.append(("build_i18n", has_i18n))
        self.sub_commands.append(("build_icons", has_icons))
        self.sub_commands.append(("build_help", has_help))
        self.sub_commands.insert(
            0, ("build_kdeui", has_kdeui)
        )  # need to run before build_py



"""distutils_extra.command.build_icons

Implement DistutilsExtra's 'build_icons' command.
"""

# Created by Sebastian Heinlein

# TODO: Address following pylint complaints
# pylint: disable=attribute-defined-outside-init,missing-function-docstring

import glob
import os

from setuptools import Command


# pylint: disable-next=invalid-name
class build_icons(Command):
    """select all icons for installation"""

    description = __doc__

    user_options = [("icon-dir=", "i", "icon directory of the source tree")]

    def initialize_options(self):
        self.icon_dir = None

    def finalize_options(self):
        if self.icon_dir is None:
            self.icon_dir = os.path.join("data", "icons")

    def run(self):
        data_files = self.distribution.data_files

        for size in glob.glob(os.path.join(self.icon_dir, "*")):
            for category in glob.glob(os.path.join(size, "*")):
                icons = []
                for icon in glob.glob(os.path.join(category, "*")):
                    if not os.path.islink(icon):
                        icons.append(icon)
                if icons:
                    data_files.append(
                        (
                            f"share/icons/hicolor/{os.path.basename(size)}"
                            f"/{os.path.basename(category)}",
                            icons,
                        )
                    )


# class build



# DistUtilsExtra.command.pylint - DistUtilsExtra command to call pylint
#
# Author: Rodney Dawes <rodney.dawes@canonical.com>
# Copyright 2009 Canonical Ltd.

"""DistUtilsExtra.command.pylint

Implements the DistUtilsExtra 'pylint' command.
"""

# TODO: Address following pylint complaints
# pylint: disable=attribute-defined-outside-init,eval-used,missing-function-docstring

import os
import subprocess

from setuptools import Command


# pylint: disable-next=invalid-name
class pylint(Command):
    """Command to run pylint and tests on a module."""

    description = "integrate pylint checks"

    user_options = [
        ("config-file=", None, "pylint config file to use"),
        ("exclude-files=", None, "list of files to exclude from lint checks"),
        ("lint-files=", None, "list of modules or packages to run lint checks on"),
    ]

    def initialize_options(self):
        self.config_file = None
        self.exclude_files = None
        self.lint_files = None

    def finalize_options(self):
        if self.config_file is None:
            self.config_file = ""
        if self.exclude_files is None:
            self.exclude_files = "[]"
        if self.lint_files is None:
            self.lint_files = f"[{self.__find_files()}]"

    def run(self):
        pylint_args = ["--output-format=parseable"]

        if self.config_file:
            pylint_args.append(f"--rcfile={self.config_file}")

        for file in eval(self.lint_files):
            pylint_args.append(file)

        with subprocess.Popen(
            ["pylint"] + pylint_args, bufsize=4096, stdout=subprocess.PIPE, text=True
        ) as process:
            output = "".join(process.stdout.readlines())

        if output != "":
            print("== Pylint notices ==")
            print(self.__group_lines_by_file(output))

    def __group_lines_by_file(self, input_):
        """Format file:line:message output as lines grouped by file."""
        outputs = []
        filename = ""
        excludes = eval(self.exclude_files)
        for line in input_.splitlines():
            current = line.split(":", 3)
            if line.startswith("    "):
                outputs.append(f"    {current[0]}")
            elif (
                line.startswith("build/") or current[0] in excludes or len(current) < 3
            ):
                pass
            elif filename == current[0]:
                outputs.append(f"    {current[1]}: {current[2]}")
            elif filename != current[0]:
                filename = current[0]
                outputs.append("")
                outputs.append(f"{filename}:")
                outputs.append(f"    {current[1]}: {current[2]}")

        return "\n".join(outputs)

    def __find_files(self):
        """Find all Python files under the current tree."""
        pyfiles = []
        for root, _, files in os.walk(os.getcwd(), topdown=False):
            for file in files:
                if file.endswith(".py"):
                    pyfiles.append(f"'{os.path.join(root, file)}'")
        pyfiles.sort()
        return ",".join(pyfiles)



"""distutils_extra.command.build_i18n

Implements the Distutils 'build_i18n' command."""

# TODO: Address following pylint complaints
# pylint: disable=attribute-defined-outside-init,broad-except,eval-used,missing-function-docstring

import glob
import os
import pathlib

from setuptools import Command


# pylint: disable-next=invalid-name,too-many-instance-attributes
class build_i18n(Command):
    """integrate the gettext framework"""

    description = __doc__

    user_options = [
        ("desktop-files=", None, ".desktop.in files that should be merged"),
        ("xml-files=", None, ".xml.in files that should be merged"),
        ("schemas-files=", None, ".schemas.in files that should be merged"),
        ("ba-files=", None, "bonobo-activation files that should be merged"),
        ("rfc822deb-files=", None, "RFC822 files that should be merged"),
        ("key-files=", None, ".key.in files that should be merged"),
        ("domain=", "d", "gettext domain"),
        ("merge-po", "m", "merge po files against template"),
        ("po-dir=", "p", "directory that holds the i18n files"),
        ("bug-contact=", None, "contact address for msgid bugs"),
    ]

    boolean_options = ["merge-po"]

    def initialize_options(self):
        self.desktop_files = []
        self.xml_files = []
        self.key_files = []
        self.schemas_files = []
        self.ba_files = []
        self.rfc822deb_files = []
        self.domain = None
        self.merge_po = False
        self.bug_contact = None
        self.po_dir = None

    def finalize_options(self):
        if self.domain is None:
            self.domain = self.distribution.metadata.name
        if self.po_dir is None:
            self.po_dir = "po"

    def run(self):
        """
        Update the language files, generate mo files and add them
        to the to be installed files
        """
        # TODO: split into smaller methods
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        if not os.path.isdir(self.po_dir):
            return

        data_files = self.distribution.data_files
        if data_files is None:
            # in case not data_files are defined in setup.py
            self.distribution.data_files = data_files = []

        if self.bug_contact is not None:
            os.environ["XGETTEXT_ARGS"] = f"--msgid-bugs-address={self.bug_contact} "

        # Print a warning if there is a Makefile that would overwrite our
        # values
        if os.path.exists(f"{self.po_dir}/Makefile"):
            self.announce(
                """
WARNING: Intltool will use the values specified from the
         existing po/Makefile in favor of the values
         from setup.cfg.
         Remove the Makefile to avoid problems."""
            )

        # If there is a po/LINGUAS file, or the LINGUAS environment variable
        # is set, only compile the languages listed there.
        selected_languages = None
        linguas_file = pathlib.Path(self.po_dir) / "LINGUAS"
        if linguas_file.is_file():
            selected_languages = linguas_file.read_text().split()
        if "LINGUAS" in os.environ:
            selected_languages = os.environ["LINGUAS"].split()

        # Update po(t) files and print a report
        # We have to change the working dir to the po dir for intltool
        cmd = ["intltool-update", (self.merge_po and "-r" or "-p"), "-g", self.domain]
        cwd = os.getcwd()
        os.chdir(self.po_dir)
        self.spawn(cmd)
        os.chdir(cwd)
        max_po_mtime = 0
        for po_file in glob.glob(f"{self.po_dir}/*.po"):
            lang = os.path.basename(po_file[:-3])
            if selected_languages and lang not in selected_languages:
                continue
            mo_dir = os.path.join("build", "mo", lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir, f"{self.domain}.mo")
            if not os.path.exists(mo_dir):
                os.makedirs(mo_dir)
            cmd = ["msgfmt", po_file, "-o", mo_file]
            po_mtime = os.path.getmtime(po_file)
            mo_mtime = os.path.exists(mo_file) and os.path.getmtime(mo_file) or 0
            if po_mtime > max_po_mtime:
                max_po_mtime = po_mtime
            if po_mtime > mo_mtime:
                self.spawn(cmd)

            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, (mo_file,)))

        # merge .in with translation
        for option, switch in (
            (self.xml_files, "-x"),
            (self.desktop_files, "-d"),
            (self.schemas_files, "-s"),
            (self.rfc822deb_files, "-r"),
            (self.ba_files, "-b"),
            (self.key_files, "-k"),
        ):
            try:
                file_set = eval(option)
            except Exception:
                continue
            for target, files in file_set:
                build_target = os.path.join("build", target)
                if not os.path.exists(build_target):
                    os.makedirs(build_target)
                files_merged = []
                for file in files:
                    if file.endswith(".in"):
                        file_merged = os.path.basename(file[:-3])
                    else:
                        file_merged = os.path.basename(file)
                    file_merged = os.path.join(build_target, file_merged)
                    cmd = ["intltool-merge", switch, self.po_dir, file, file_merged]
                    mtime_merged = (
                        os.path.exists(file_merged)
                        and os.path.getmtime(file_merged)
                        or 0
                    )
                    mtime_file = os.path.getmtime(file)
                    if mtime_merged < max_po_mtime or mtime_merged < mtime_file:
                        # Only build if output is older than input (.po,.in)
                        self.spawn(cmd)
                    files_merged.append(file_merged)
                data_files.append((target, files_merged))


# class build



"""Implement the Distutils "build_help" command."""

# TODO: Address following pylint complaints
# pylint: disable=attribute-defined-outside-init,missing-function-docstring

import os.path
from glob import glob

from setuptools import Command


# pylint: disable-next=invalid-name
class build_help(Command):
    """install Mallard or DocBook XML based documentation"""

    description = __doc__
    user_options = [("help-dir", None, "help directory in the source tree")]

    def initialize_options(self):
        self.help_dir = None

    def finalize_options(self):
        if self.help_dir is None:
            self.help_dir = "help"

    def get_data_files(self):
        data_files = []
        name = self.distribution.metadata.name

        for path in glob(os.path.join(self.help_dir, "*")):
            lang = os.path.basename(path)
            path_xml = os.path.join("share/help", lang, name)
            path_figures = os.path.join("share/help", lang, name, "figures")

            docbook_files = glob(f"{path}/index.docbook")
            docbook_files_extra = glob(f"{path}/*.xml")
            mallard_files = glob(f"{path}/*.page")
            data_files.append(
                (path_xml, docbook_files + docbook_files_extra + mallard_files)
            )
            data_files.append((path_figures, glob(f"{path}/figures/*.png")))

        return data_files

    def run(self):
        self.announce("Setting up help files...")

        data_files = self.distribution.data_files
        data_files.extend(self.get_data_files())



#!/usr/bin/env python3

import pathlib
import re
import sys

from setuptools import setup

sys.path.insert(0, ".")
from DistUtilsExtra import __version__ as pkgversion


def get_debian_version() -> str:
    """Look what Debian version we have."""
    changelog = pathlib.Path(__file__).parent / "debian" / "changelog"
    with changelog.open("r", encoding="utf-8") as changelog_f:
        head = changelog_f.readline()
    match = re.compile(r".*\((.*)\).*").match(head)
    if not match:
        raise ValueError(f"Failed to extract Debian version from '{head}'.")
    return match.group(1)


def _check_debian_version() -> None:
    debian_version = get_debian_version()
    if not debian_version.startswith(pkgversion):
        print(
            f"Error: Debian version '{debian_version}' does not"
            f" start with DistUtilsExtra.__version__ '{pkgversion}'.",
            file=sys.stderr,
        )
        sys.exit(1)


_check_debian_version()

setup(
    name="python-distutils-extra",
    version=pkgversion,
    author="Sebastian Heinlein, Martin Pitt",
    author_email="sebi@glatzor.de, martin.pitt@ubuntu.com",
    description="Add support for i18n, documentation and icons to distutils",
    packages=["DistUtilsExtra", "DistUtilsExtra.command"],
    license="GNU GPL",
    platforms="posix",
    entry_points={
        "distutils.commands": [
            "build = DistUtilsExtra.command.build_extra:build_extra",
            "build_i18n = DistUtilsExtra.command.build_i18n:build_i18n",
            "build_icons = DistUtilsExtra.command.build_icons:build_icons",
            "build_help = DistUtilsExtra.command.build_help:build_help",
            "clean_i18n = DistUtilsExtra.command.clean_i18n:clean_i18n",
            "pylint = DistUtilsExtra.command.pylint:pylint",
        ]
    },
)



#!/usr/bin/python3

"""Test DistUtilsExtra.auto."""

# TODO: Address following pylint complaints
# pylint: disable=consider-using-with,invalid-name,too-many-lines,use-a-generator

import os
import pathlib
import re
import shutil
import subprocess
import tempfile
import unittest


# pylint: disable-next=too-many-public-methods
class T(unittest.TestCase):
    """Test DistUtilsExtra.auto."""

    def setUp(self):
        self.maxDiff = None
        self.src = tempfile.mkdtemp()

        self._mksrc(
            "setup.py",
            """
# ignore warning about import from local path
import warnings
warnings.filterwarnings('ignore', 'Module DistUtilsExtra was already imported from.*')
warnings.filterwarnings('ignore', 'pipe2 set errno ENOSYS.*')
warnings.filterwarnings('ignore', 'setup.py install is deprecated.*')

from DistUtilsExtra.auto import setup

setup(
    name='foo',
    version='0.1',
    description='Test suite package',
    url='https://foo.example.com',
    license='GPL v2 or later',
    author='Martin Pitt',
    author_email='martin.pitt@example.com',
)
""",
        )
        self.snapshot = None
        self.install_tree = None

    def tearDown(self):
        try:
            # check that setup.py clean removes everything
            (o, e, s) = self.setup_py(["clean", "-a"])
            self.assertEqual(s, 0, o + e)
            cruft = self.diff_snapshot()
            self.assertEqual(cruft, "", f"no cruft after cleaning:\n{cruft}")
        finally:
            shutil.rmtree(self.src)
            if self.snapshot:
                shutil.rmtree(self.snapshot)
            if self.install_tree:
                shutil.rmtree(self.install_tree)
            self.src = None
            self.snapshot = None
            self.install_tree = None

    def assert_egg_info_directory_is_present_and_well(self):
        """Check that no .egg-info file is present, that an egg_info directory
        is present and that it contains the expected files"""

        f = self.installed_files()
        # All files are in an .egg-info directory; no .egg-info file is created
        self.assertFalse(any([_.endswith(".egg-info") for _ in f]))
        # There are 4 files in said directory
        self.assertEqual(len(f), 4)
        # Check that the four exist
        self.assertTrue(
            all(
                [
                    any(
                        [
                            _.endswith(c)
                            for c in [
                                "PKG-INFO",
                                "SOURCES.txt",
                                "dependency_links.txt",
                                "top_level.txt",
                            ]
                        ]
                    )
                    for _ in f
                ]
            )
        )

    #
    # actual tests come here
    #

    def test_empty(self):
        """empty source tree (just setup.py)"""

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        self.assert_egg_info_directory_is_present_and_well()

    def test_vcs(self):
        """Ignores revision control files"""

        self._mksrc(".shelf/1")
        self._mksrc(".bzr/revs")
        self._mksrc(".git/config")
        self._mksrc(".svn/revs")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        self.assert_egg_info_directory_is_present_and_well()

    def test_modules(self):
        """Python modules"""

        self._mksrc("yesme.py", b'x ="a\xc3\xa4b\xe2\x99\xa5"'.decode("UTF-8"))
        self._mksrc("stuff/notme.py", b'x ="a\xc3\xa4b\xe2\x99\xa5"'.decode("UTF-8"))
        self._mksrc(
            "stuff/withencoding.py",
            b"# -*- Mode: Python; coding: utf-8; -*- \nfoo = 1".decode("UTF-8"),
        )

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  stuff/notme.py\n", o)

        f = "\n".join(self.installed_files())
        self.assertIn("-packages/yesme.py", f)
        self.assertNotIn("notme", f)

    def test_packages(self):
        """Python packages"""

        self._mksrc("foopkg/__init__.py", "")
        self._mksrc("foopkg/bar.py")
        self._mksrc("foopkg/baz.py")
        self._mksrc("noinit/notme.py")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  noinit/notme.py\n", o)

        f = "\n".join(self.installed_files())
        self.assertIn("foopkg/__init__.py", f)
        self.assertIn("foopkg/bar.py", f)
        self.assertNotIn("noinit", f)

    def test_dbus(self):
        """D-BUS configuration and service files"""

        # D-BUS ACL configuration file
        self._mksrc(
            "daemon/com.example.foo.conf",
            """<!DOCTYPE busconfig PUBLIC
 "-//freedesktop//DTD D-BUS Bus Configuration 1.0//EN"
 "http://www.freedesktop.org/standards/dbus/1.0/busconfig.dtd">
<busconfig>
</busconfig>""",
        )

        # non-D-BUS configuration file
        self._mksrc("daemon/defaults.conf", "start = True\nlog = syslog")

        # D-BUS system service
        self._mksrc(
            "daemon/com.example.foo.service",
            """[D-BUS Service]
Name=com.example.Foo
Exec=/usr/lib/foo/foo_daemon
User=root""",
        )

        # D-BUS session service
        self._mksrc(
            "gui/com.example.foo.gui.service",
            """[D-BUS Service]
Name=com.example.Foo.GUI
Exec=/usr/bin/foo-gtk
""",
        )

        # non-D-BUS .service file
        self._mksrc("stuff/super.service", "I am a file")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  stuff/super.service\n", o)

        f = self.installed_files()
        self.assertEqual(len(f), 7)  # 3 D-BUS files plus 4 files in egg-info directory
        self.assertIn("/etc/dbus-1/system.d/com.example.foo.conf", f)
        self.assertIn("/usr/share/dbus-1/system-services/com.example.foo.service", f)
        self.assertIn("/usr/share/dbus-1/services/com.example.foo.gui.service", f)
        self.assertNotIn("super.service", "\n".join(f))

    def test_gsettings(self):
        """GSettings schema files"""

        # schema files in dedicated directory
        self._mksrc("data/glib-2.0/schemas/org.test.myapp.gschema.xml")
        self._mksrc("data/glib-2.0/schemas/gschemas.compiled")
        # schema files in data directory
        self._mksrc("data/org.test.myapp2.gschema.xml")
        self._mksrc("data/gschemas.compiled")

        (_, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)

        f = self.installed_files()
        self.assertEqual(
            len(f), 6
        )  # 2 schema files plus 4 files in .egg-info directory
        self.assertIn("/usr/share/glib-2.0/schemas/org.test.myapp.gschema.xml", f)
        self.assertNotIn("gschemas.compiled", "\n".join(f))

    def test_apport_hook(self):
        """Apport hooks"""

        self._mksrc(
            "apport/foo.py",
            """import os
def add_info(report):
    pass
""",
        )

        self._mksrc(
            "apport/source_foo.py",
            """import os
def add_info(report):
    pass
""",
        )

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertEqual(len(f), 6, f)  # 2 hook files plus 4 in .egg-info
        self.assertIn("/usr/share/apport/package-hooks/foo.py", f)
        self.assertIn("/usr/share/apport/package-hooks/source_foo.py", f)

    def test_po(self):
        """gettext *.po files"""

        self._mkpo()

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)
        f = self.installed_files()
        self.assertIn("/usr/share/locale/de/LC_MESSAGES/foo.mo", f)
        self.assertIn("/usr/share/locale/fr/LC_MESSAGES/foo.mo", f)
        self.assertNotIn("junk", "\n".join(f))

        msgunfmt = subprocess.Popen(
            [
                "msgunfmt",
                os.path.join(
                    self.install_tree, "usr/share/locale/de/LC_MESSAGES/foo.mo"
                ),
            ],
            stdout=subprocess.PIPE,
        )
        out = msgunfmt.communicate()[0].decode()
        self.assertEqual(out, self._src_contents("po/de.po"))

    def test_policykit(self):
        """*.policy.in PolicyKit files"""

        self._mksrc(
            "daemon/com.example.foo.policy.in",
            """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE policyconfig PUBLIC
 "-//freedesktop//DTD PolicyKit Policy Configuration 1.0//EN"
 "http://www.freedesktop.org/standards/PolicyKit/1.0/policyconfig.dtd">
<policyconfig>
  <vendor>Foo project</vendor>
  <vendor_url>https://foo.example.com</vendor_url>

  <action id="com.example.foo.greet">
    <_description>Good morning</_description>
    <_message>Hello</_message>
    <defaults>
      <allow_active>yes</allow_active>
    </defaults>
  </action>
</policyconfig>""",
        )

        self._mkpo()
        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertIn("/usr/share/polkit-1/actions/com.example.foo.policy", f)
        p = self._installed_contents(
            "usr/share/polkit-1/actions/com.example.foo.policy"
        )
        self.assertIn("<description>Good morning</description>", p)
        self.assertIn('<description xml:lang="de">Guten Morgen</description>', p)
        self.assertIn("<message>Hello</message>", p)
        self.assertIn('<message xml:lang="de">Hallo</message>', p)

    def test_desktop(self):
        """*.desktop.in files"""

        self._mksrc(
            "gui/foogtk.desktop.in",
            """[Desktop Entry]
_Name=Hello
_Comment=Good morning
Exec=/bin/foo""",
        )
        self._mksrc(
            "gui/autostart/fooapplet.desktop.in",
            """[Desktop Entry]
_Name=Hello
_Comment=Good morning
Exec=/usr/bin/fooapplet""",
        )
        self._mkpo()
        self._mksrc(
            "data/foosettings.desktop.in",
            """[Desktop Entry]
_Name=Hello
_Comment=Good morning
Exec=/bin/foosettings""",
        )

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertIn("/usr/share/autostart/fooapplet.desktop", f)
        self.assertIn("/usr/share/applications/foogtk.desktop", f)
        self.assertIn("/usr/share/applications/foosettings.desktop", f)
        # data/*.desktop.in shouldn't go to data dir
        self.assertNotIn("/usr/share/foo/", f)

        p = self._installed_contents("usr/share/autostart/fooapplet.desktop")
        self.assertIn("\nName=Hello\n", p)
        self.assertIn("\nName[de]=Hallo\n", p)
        self.assertIn("\nComment[fr]=Bonjour\n", p)

    def test_icons(self):
        """data/icons/"""

        self._mksrc("data/icons/scalable/actions/press.png")
        self._mksrc("data/icons/48x48/apps/foo.png")
        scalable_icon_path = os.path.join(self.src, "data", "icons", "scalable")
        os.symlink(
            os.path.join(scalable_icon_path, "actions", "press.png"),
            os.path.join(scalable_icon_path, "actions", "crunch.png"),
        )

        # test broken symlink, too
        os.mkdir(os.path.join(scalable_icon_path, "mimetypes"))
        os.symlink(
            "../apps/foo.svg",
            os.path.join(scalable_icon_path, "mimetypes", "text-x-foo.svg"),
        )

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertIn("/usr/share/icons/hicolor/scalable/actions/press.png", f)
        self.assertIn("/usr/share/icons/hicolor/scalable/actions/crunch.png", f)
        self.assertIn("/usr/share/icons/hicolor/48x48/apps/foo.png", f)
        self.assertTrue(
            os.path.islink(
                os.path.join(
                    self.install_tree,
                    "usr/share/icons/hicolor/scalable/actions/crunch.png",
                )
            )
        )
        self.assertTrue(
            os.path.islink(
                os.path.join(
                    self.install_tree,
                    "usr/share/icons/hicolor/scalable/mimetypes/text-x-foo.svg",
                )
            )
        )

    def test_data(self):
        """Auxiliary files in data/"""

        # have some explicitly covered files, to check that they don't get
        # installed into prefix/share/foo/ again
        self._mksrc(
            "setup.py",
            """
import warnings
from DistUtilsExtra.auto import setup
from glob import glob

warnings.filterwarnings('ignore', 'setup.py install is deprecated.*')

setup(
    name='foo',
    version='0.1',
    description='Test suite package',
    url='https://foo.example.com',
    license='GPL v2 or later',
    author='Martin Pitt',
    author_email='martin.pitt@example.com',

    data_files = [
      ('/lib/udev/rules.d', ['data/40-foo.rules']),
      ('/etc/foo', glob('data/*.conf')),
    ]
)
""",
        )

        self._mksrc("data/stuff")
        self._mksrc("data/handlers/red.py", 'import sys\nprint ("RED")')
        self._mksrc("data/handlers/blue.py", 'import sys\nprint ("BLUE")')
        self._mksrc("data/40-foo.rules")
        self._mksrc("data/blob1.conf")
        self._mksrc("data/blob2.conf")
        os.symlink("stuff", os.path.join(self.src, "data", "stufflink"))

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertIn("/usr/share/foo/stuff", f)
        self.assertIn("/usr/share/foo/stufflink", f)
        self.assertTrue(
            os.path.islink(
                os.path.join(self.install_tree, "usr", "share", "foo", "stufflink")
            )
        )
        self.assertIn("/usr/share/foo/handlers/red.py", f)
        self.assertIn("/usr/share/foo/handlers/blue.py", f)
        self.assertIn("/lib/udev/rules.d/40-foo.rules", f)
        self.assertIn("/etc/foo/blob1.conf", f)
        self.assertIn("/etc/foo/blob2.conf", f)
        self.assertNotIn("/usr/share/foo/blob1.conf", f)
        self.assertNotIn("/usr/share/foo/40-foo.rules", f)

    def test_scripts(self):
        """scripts"""

        # these should get autoinstalled
        self._mksrc("bin/yell", "#!/bin/sh", True)
        self._mksrc("bin/shout", "#!/bin/sh", True)
        self._mksrc(
            "bin/foo", b"#!/usr/bin/python\n# \xc2\xa9 copyright".decode("UTF-8"), True
        )
        os.symlink("shout", os.path.join(self.src, "bin", "shoutlink"))

        # these shouldn't
        self._mksrc("daemon/food", "#!/bin/sh", True)  # not in bin/
        self._mksrc("foob", "#!/bin/sh", True)  # not named like project
        # not executable
        self._mksrc(
            "bin/whisper", b"#!/usr/bin/python\n# \xc2\xa9 copyright".decode("UTF-8")
        )

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  foob", o)
        self.assertIn("\n  bin/whisper", o)
        self.assertIn("\n  daemon/food", o)

        f = self.installed_files()
        self.assertIn("/usr/bin/yell", f)
        self.assertIn("/usr/bin/shout", f)
        self.assertIn("/usr/bin/shoutlink", f)
        self.assertTrue(
            os.path.islink(os.path.join(self.install_tree, "usr", "bin", "shoutlink"))
        )
        self.assertIn("/usr/bin/foo", f)
        ftext = "\n".join(f)
        self.assertNotIn("food", ftext)
        self.assertNotIn("foob", ftext)
        self.assertNotIn("whisper", ftext)

        # verify that they are executable
        binpath = os.path.join(self.install_tree, "usr", "bin")
        self.assertTrue(os.access(os.path.join(binpath, "yell"), os.X_OK))
        self.assertTrue(os.access(os.path.join(binpath, "shout"), os.X_OK))
        self.assertTrue(os.access(os.path.join(binpath, "foo"), os.X_OK))

    def test_pot_manual(self):
        """PO template creation with manual POTFILES.in"""

        self._mk_i18n_source()
        self._mksrc("po/foo.pot", "")
        # only do a subset here
        self._mksrc(
            "po/POTFILES.in",
            """
gtk/main.py
gui/foo.desktop.in
[type: gettext/glade]gtk/test.ui""",
        )

        (o, e, s) = self.setup_py(["build"])
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        # POT file should not be shown as not recognized
        self.assertNotIn("\n  po/foo.pot\n", o)

        pot = self._src_contents("po/foo.pot")

        self.assertNotIn('msgid "no"', pot)
        self.assertIn('msgid "yes1"', pot)
        self.assertIn('msgid "yes2 %s"', pot)
        self.assertNotIn('msgid "yes5"', pot)  # we didn't add helpers.py
        self.assertIn('msgid "yes7"', pot)  # we did include the desktop file
        self.assertNotIn('msgid "yes5"', pot)  # we didn't add helpers.py
        self.assertIn('msgid "yes11"', pot)  # we added one GTKBuilder file
        self.assertNotIn('msgid "yes12"', pot)  # ... but not the other

    def test_pot_auto(self):
        """PO template creation with automatic POTFILES.in"""

        self._mk_i18n_source()

        (o, e, s) = self.setup_py(["build"])
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        # POT file should not be shown as not recognized
        self.assertNotIn("\n  po/foo.pot\n", o)

        pot = self._src_contents("po/foo.pot")

        self.assertNotIn('msgid "no"', pot)
        for i in range(2, 15):
            self.assertTrue(
                f'msgid "yes{int(i)}' in pot or f'msgid ""\n"yes{int(i)}' in pot,
                f"yes{int(i)}",
            )
        # above loop would match yes11 to yes1 as well, so test it explicitly
        self.assertIn('msgid "yes1"', pot)

    def test_pot_auto_explicit(self):
        """PO template creation with automatic POTFILES.in and explicit scripts"""

        self._mk_i18n_source()

        # add some additional binaries here which aren't caught by default
        self._mksrc("cli/client-cli", "#!/usr/bin/python\nprint (_('yes15'))", True)
        self._mksrc("gtk/client-gtk", '#!/usr/bin/python\nprint (_("yes16"))', True)
        # this is the most tricky case: intltool doesn't consider them Python
        # files by default and thus just looks for _(""):
        self._mksrc("kde/client-kde", "#!/usr/bin/python\nprint (_('yes17'))", True)
        self._mksrc("po/POTFILES.in.in", "gtk/client-gtk\nkde/client-kde")
        self._mksrc(
            "setup.py",
            """
from DistUtilsExtra.auto import setup

import warnings
warnings.filterwarnings('ignore', 'pipe2 set errno ENOSYS.*')

setup(
    name='foo',
    version='0.1',
    data_files=[('share/foo', ['gtk/client-gtk', 'kde/client-kde'])],
    scripts=['cli/client-cli'],
)
""",
        )

        (o, e, s) = self.setup_py(["build"])
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        # POT file should not be shown as not recognized
        self.assertNotIn("\n  po/foo.pot\n", o)

        pot = self._src_contents("po/foo.pot")

        self.assertNotIn('msgid "no"', pot)
        for i in range(2, 18):
            self.assertTrue(
                f'msgid "yes{int(i)}' in pot or f'msgid ""\n"yes{int(i)}' in pot,
                f"yes{int(i)}",
            )
        # above loop would match yes11 to yes1 as well, so test it explicitly
        self.assertIn('msgid "yes1"', pot)

    def test_standard_files(self):
        """Standard files (MANIFEST.in, COPYING, etc.)"""

        self._mksrc("AUTHORS")
        self._mksrc("COPYING")
        self._mksrc("LICENSE")
        self._mksrc("COPYING.LIB")
        self._mksrc("README.txt")
        self._mksrc("MANIFEST.in", content="# dummy")
        self._mksrc("MANIFEST")
        self._mksrc("NEWS")
        self._mksrc("TODO")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertIn("/usr/share/doc/foo/README.txt", f)
        self.assertIn("/usr/share/doc/foo/NEWS", f)
        ftext = "\n".join(f)
        self.assertNotIn("MANIFEST", ftext)
        self.assertNotIn("COPYING", ftext)
        self.assertNotIn("COPYING", ftext)
        self.assertNotIn("AUTHORS", ftext)
        self.assertNotIn("TODO", ftext)

        # sub-dir READMEs shouldn't be installed by default
        self.snapshot = None
        self._mksrc("extra/README")
        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  extra/README\n", o)

    def test_sdist(self):
        """default MANIFEST"""

        good = [
            "AUTHORS",
            "README.txt",
            "COPYING",
            "helpers.py",
            "foo/__init__.py",
            "foo/bar.py",
            "tests/all.py",
            "gui/x.desktop.in",
            "backend/foo.policy.in",
            "daemon/backend.conf",
            "x/y",
            "po/de.po",
            "po/foo.pot",
            ".quickly",
            "data/icons/16x16/apps/foo.png",
            "bin/foo",
            "backend/food",
            "backend/com.example.foo.service",
            "gtk/main.glade",
            "dist/extra.tar.gz",
        ]
        bad = [
            "po/de.mo",
            ".helpers.py.swp",
            ".bzr/index",
            ".svn/index",
            ".git/index",
            "bin/foo~",
            "backend/foo.pyc",
            "dist/foo-0.2.tar.gz",
            ".shelf/1",
            ".bzr/revs",
            ".git/config",
        ]

        for f in good + bad:
            self._mksrc(f)

        (_, e, s) = self.setup_py(["sdist"])
        self.assertEqual(e, "")
        self.assertEqual(s, 0)

        tarball = pathlib.Path(self.src) / "dist" / "foo-0.1.tar.gz"
        tar = subprocess.run(
            ["tar", "tf", str(tarball)], capture_output=True, text=True, check=True
        )
        tarball.unlink()

        manifest = [re.sub(r"^foo-0\.1/", "", f) for f in tar.stdout.splitlines()]

        for f in good:
            self.assertIn(f, manifest)
        for f in bad:
            self.assertNotIn(f, manifest)

    def test_ui(self):
        """GtkBuilder/Qt *.ui"""

        self._mksrc(
            "gtk/test.ui",
            b"""<?xml version="1.0"?>
<interface>
  <requires lib="gtk+" version="2.16"/>
  <object class="GtkWindow" id="window1">
    <property name="title" translatable="yes">my\xe2\x99\xa5</property>
    <child><placeholder/></child>
  </object>
</interface>""".decode(
                "UTF-8"
            ),
        )

        self._mksrc(
            "gtk/settings.ui",
            """<?xml version="1.0"?>
<!-- Generated with glade 3.18.3 -->
<interface domain="foobar">
  <requires lib="gtk+" version="2.16"/>
  <object class="GtkWindow" id="window2">
    <property name="title" translatable="yes">yes12</property>
    <child><placeholder/></child>
  </object>
</interface>""",
        )

        self._mksrc(
            "kde/mainwindow.ui",
            """<?xml version="1.0"?>
<ui version="4.0">
 <class>CrashDialog</class>
 <widget class="QDialog" name="CrashDialog">
 </widget>
</ui>
""",
        )

        self._mksrc("someweird.ui")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  someweird.ui\n", o)

        f = self.installed_files()
        self.assertIn("/usr/share/foo/test.ui", f)
        self.assertIn("/usr/share/foo/settings.ui", f)
        self.assertIn("/usr/share/foo/mainwindow.ui", f)
        ftext = "\n".join(f)
        self.assertNotIn("someweird", ftext)

    def test_manpages(self):
        """manpages"""

        self._mksrc("man/foo.1", '.TH foo 1 "Jan 01, 1900" "Joe Developer"')
        self._mksrc(
            "daemon/food.8",
            '." some comment\n.TH food 8 "Jan 01, 1900" "Joe Developer"',
        )
        self._mksrc("cruft/food.1", "")
        self._mksrc("daemon/notme.s", '.TH food 8 "Jan 01, 1900" "Joe Developer"')

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  cruft/food.1\n", o)
        self.assertIn("\n  daemon/notme.s\n", o)

        f = self.installed_files()
        self.assertIn("/usr/share/man/man1/foo.1", f)
        self.assertIn("/usr/share/man/man8/food.8", f)
        ftext = "\n".join(f)
        self.assertNotIn("food.1", ftext)
        self.assertNotIn("notme", ftext)

    def test_etc(self):
        """etc/*"""

        self._mksrc("etc/cron.daily/foo")
        self._mksrc("etc/foo.conf")
        self._mksrc("etc/init.d/foo", executable=True)
        d = os.path.join(self.src, "etc", "cron.weekly")
        os.mkdir(d)
        os.symlink(os.path.join("..", "cron.daily", "foo"), os.path.join(d, "foo"))

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

        f = self.installed_files()
        self.assertIn("/etc/cron.daily/foo", f)
        self.assertIn("/etc/cron.weekly/foo", f)
        self.assertIn("/etc/init.d/foo", f)
        self.assertIn("/etc/foo.conf", f)

        # verify that init script is executable
        self.assertTrue(
            os.access(os.path.join(self.install_tree, "etc", "init.d", "foo"), os.X_OK)
        )
        # verify that symlinks get preserved
        self.assertTrue(
            os.path.islink(os.path.join(self.install_tree, "etc", "cron.weekly", "foo"))
        )

        # check that we can install again into the same source tree
        (o, e, s) = self.setup_py(
            ["install", "--no-compile", "--prefix=/usr", f"--root={self.install_tree}"]
        )
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertNotIn("following files are not recognized", o)

    def test_requires_provides(self):
        """automatic requires/provides"""

        for needed_pkg in ["pkg_resources", "httplib2", "gi.repository.GLib"]:
            try:
                __import__(needed_pkg)
            except ImportError:
                self.fail(
                    f"You need to have {needed_pkg} installed"
                    f" for this test suite to work"
                )

        self._mksrc("foo/__init__.py", "")
        self._mksrc(
            "foo/stuff.py",
            """import xml.parsers.expat
import os, os.path, email.mime, setuptools.command.sdist
from email import header as h
import httplib2.iri2uri, unknown
from . bar import poke
from bar.poke import x
import grab_cli
import broken
""",
        )

        self._mksrc("foo/bar/__init__.py", "")
        self._mksrc("foo/bar/poke.py", "from . import broken\ndef x(): pass")
        self._mksrc(
            "foo/bar/broken.py", 'raise RuntimeError("cannot initialize system")'
        )

        self._mksrc("mymod.py", "import foo\nfrom foo.bar.poke import x")
        # trying to import this will cause setup.py to not process any args any more
        self._mksrc(
            "grab_cli.py",
            "from optparse import OptionParser\nOptionParser().parse_args()",
        )
        # trying to import this will break setup.py
        self._mksrc("broken.py", 'raise SystemError("cannot initialize system")')
        self._mksrc(
            "pygi.py", "from gi.repository import GLib\nimport gi.repository.GObject"
        )

        self._mksrc(
            "bin/foo-cli",
            """#!/usr/bin/python
import sys
import pkg_resources
import foo.bar
from httplib2 import iri2uri

print ('import iamnota.module')
""",
            executable=True,
        )

        # this shouldn't be treated specially
        self._mksrc("data/example-code/template.py", "import example.module")
        self._mksrc("data/example-code/mymod/__init__.py", "")
        self._mksrc("data/example-code/mymod/shiny.py", "import example.othermod")

        (o, e, s) = self.do_install()
        self.assertEqual(s, 0, e)
        self.assertEqual(e, "ERROR: Python module unknown not found\n")
        self.assertNotIn("following files are not recognized", o)

        inst = self.installed_files()
        self.assertIn("/usr/share/foo/example-code/template.py", inst)
        self.assertIn("/usr/share/foo/example-code/mymod/shiny.py", inst)
        for f in inst:
            if "template.py" in f or "shiny" in f:
                self.assertNotIn("packages", f)

        # parse .egg-info directory
        (o, e, s) = self.setup_py(["install_egg_info", "-d", self.install_tree])
        self.assertEqual(e, "ERROR: Python module unknown not found\n")
        in_egg_paths = [x for x in inst if ".egg-info/" in x]
        self.assertEqual(len(in_egg_paths), 4)  # Always 4 files in .egg-info directory

        pkginfo = self._installed_contents(
            [x for x in in_egg_paths if x.endswith("PKG-INFO")][0].strip(os.path.sep)
        ).splitlines()
        self.assertIn("Name: foo", pkginfo)

        # check provides
        prov = [
            prop.split(" ", 1)[1] for prop in pkginfo if prop.startswith("Provides: ")
        ]
        self.assertEqual(set(prov), set(["foo", "mymod", "broken", "grab_cli", "pygi"]))

        # check requires
        req = [
            prop.split(" ", 1)[1] for prop in pkginfo if prop.startswith("Requires: ")
        ]
        self.assertEqual(
            set(req),
            set(
                [
                    "httplib2",
                    "pkg_resources",
                    "gi.repository.GLib",
                    "gi.repository.GObject",
                    "setuptools.command.sdist",
                ]
            ),
        )

    def test_help_docbook(self):
        """Docbook XML help"""

        self._mksrc("help/C/index.docbook")
        self._mksrc("help/C/legal.xml")
        self._mksrc("help/C/figures/mainscreen.png")
        self._mksrc("help/de/index.docbook")
        self._mksrc("help/de/legal.xml")
        self._mksrc("help/de/figures/mainscreen.png")

        self._mksrc("help/weird.xml")
        self._mksrc("help/notme.png")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  help/weird.xml\n", o)
        self.assertIn("\n  help/notme.png\n", o)

        f = self.installed_files()
        self.assertIn("/usr/share/help/C/foo/index.docbook", f)
        self.assertIn("/usr/share/help/C/foo/legal.xml", f)
        self.assertIn("/usr/share/help/C/foo/figures/mainscreen.png", f)
        self.assertIn("/usr/share/help/de/foo/index.docbook", f)
        self.assertIn("/usr/share/help/de/foo/legal.xml", f)
        self.assertIn("/usr/share/help/de/foo/figures/mainscreen.png", f)

    def test_help_mallard(self):
        """Mallard XML help"""

        self._mksrc("help/C/index.page")
        self._mksrc("help/C/legal.page")
        self._mksrc("help/C/figures/mainscreen.png")
        self._mksrc("help/de/index.page")
        self._mksrc("help/de/legal.page")
        self._mksrc("help/de/figures/mainscreen.png")

        self._mksrc("help/weird.page")
        self._mksrc("help/notme.png")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  help/weird.page\n", o)
        self.assertIn("\n  help/notme.png\n", o)

        f = self.installed_files()
        self.assertIn("/usr/share/help/C/foo/index.page", f)
        self.assertIn("/usr/share/help/C/foo/legal.page", f)
        self.assertIn("/usr/share/help/C/foo/figures/mainscreen.png", f)
        self.assertIn("/usr/share/help/de/foo/index.page", f)
        self.assertIn("/usr/share/help/de/foo/legal.page", f)
        self.assertIn("/usr/share/help/de/foo/figures/mainscreen.png", f)

    def test_binary_files(self):
        """Binary files are ignored"""

        with open(os.path.join(self.src, "binary_trap"), "wb") as f:
            f.write(b"\x00\x01abc\xFF\xFE")
        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)
        self.assertIn("following files are not recognized", o)
        self.assertIn("\n  binary_trap\n", o)

        self.assert_egg_info_directory_is_present_and_well()

    def test_utf8_filenames(self):
        """UTF-8 file names"""

        bin_fname = b"a\xc3\xa4b.bin".decode("UTF-8")
        with open(os.path.join(self.src, bin_fname).encode("UTF-8"), "wb") as f:
            f.write(b"\x00\x01abc\xFF\xFE")

        (o, e, s) = self.do_install()
        self.assertEqual(e, "")
        self.assertEqual(s, 0)

        self.assert_egg_info_directory_is_present_and_well()

        self.assertIn("following files are not recognized", o)
        # this might not be the correct file name when the locale is e. g. C
        self.assertIn("b.bin\n", o)

    #
    # helper methods
    #

    def setup_py(self, args):
        """Run setup.py with given arguments.

        For convenience, this snapshots the tree if no snapshot exists yet.

        Return (out, err, exitcode) triple.
        """
        if not self.snapshot:
            self.do_snapshot()

        env = os.environ.copy()
        oldcwd = os.getcwd()
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] = oldcwd + os.pathsep + env["PYTHONPATH"]
        else:
            env["PYTHONPATH"] = oldcwd
        # unset envvars that alter results
        env.pop("LINGUAS", "")
        env.pop("PYTHONDONTWRITEBYTECODE", "")
        os.chdir(self.src)
        s = subprocess.Popen(
            ["/proc/self/exe", "setup.py"] + args,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        (out, err) = s.communicate()
        out = out.decode()
        err = err.decode()
        os.chdir(oldcwd)

        return (out, err, s.returncode)

    def do_install(self):
        """Run setup.py install into temporary tree.

        Return (out, err, exitcode) triple.
        """
        self.install_tree = tempfile.mkdtemp()

        self.setup_py(["build"])
        return self.setup_py(
            [
                "install",
                "--no-compile",
                "--skip-build",
                "--prefix=/usr",
                "--install-data=/usr",
                "--install-scripts=/usr/bin",
                f"--root={self.install_tree}",
            ]
        )

    def installed_files(self):
        """Return list of file paths in install tree."""

        result = []
        for root, _, files in os.walk(self.install_tree):
            assert root.startswith(self.install_tree)
            r = root[len(self.install_tree) :]
            for f in files:
                result.append(os.path.join(r, f))
        return result

    def _mksrc(self, path, content=None, executable=False):
        """Create a file in the test source tree."""

        path = os.path.join(self.src, path)
        directory = os.path.dirname(path)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        with open(path, "wb") as f:
            if content is None:
                # default content, to spot with diff
                f.write(b"dummy")
            else:
                f.write(f"{content}\n".encode("UTF-8"))

        if executable:
            os.chmod(path, 0o755)

    def do_snapshot(self):
        """Snapshot source tree.

        This should be called after a test set up all source files.
        """
        assert self.snapshot is None, "snapshot already taken"

        self.snapshot = tempfile.mkdtemp()
        shutil.copytree(self.src, os.path.join(self.snapshot, "s"), symlinks=True)

    def diff_snapshot(self):
        """Compare source tree to snapshot, excluding known offenders.

        Check https://github.com/pypa/setuptools/issues/1347 for reference

        Return diff -Nur output.
        """
        assert self.snapshot, "no snapshot taken"
        diff = subprocess.run(
            [
                "diff",
                "-x",
                "foo.pot",
                "-x",
                "*.pyc",
                "-x",
                "*.egg-info",
                "-Nur",
                os.path.join(self.snapshot, "s"),
                self.src,
            ],
            capture_output=True,
            check=False,
            text=True,
        )
        return diff.stdout

    def _mkpo(self):
        """Create some example po files."""

        self._mksrc("po/POTFILES.in", "")
        self._mksrc(
            "po/de.po",
            '''msgid ""
msgstr "Content-Type: text/plain; charset=UTF-8\\n"

msgid "Good morning"
msgstr "Guten Morgen"

msgid "Hello"
msgstr "Hallo"''',
        )
        self._mksrc(
            "po/fr.po",
            '''msgid ""
msgstr "Content-Type: text/plain; charset=UTF-8\\n"

msgid "Good morning"
msgstr "Bonjour"''',
        )

    def _mk_i18n_source(self):
        """Create some example source files with gettext calls"""

        self._mksrc(
            "gtk/main.py",
            """print (_("yes1"))
print ("no1")
print (__("no2"))
x = _('yes2 %s') % y

def f():
    print (_("yes3"))
    return _('yes6')""",
        )

        self._mksrc(
            "helpers.py",
            '''
print (f(_("yes4")))
print (_(\'\'\'yes5
even more
lines\'\'\'))
print (_("""yes6
more lines"""))
print (\'\'\'no3
boo\'\'\')
print ("""no4
more""")''',
        )

        self._mksrc(
            "gui/foo.desktop.in",
            """[Desktop Entry]
_Name=yes7
_Comment=yes8
Icon=no5
Exec=/usr/bin/foo""",
        )

        self._mksrc(
            "daemon/com.example.foo.policy.in",
            """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE policyconfig PUBLIC
 "-//freedesktop//DTD PolicyKit Policy Configuration 1.0//EN"
 "http://www.freedesktop.org/standards/PolicyKit/1.0/policyconfig.dtd">
<policyconfig>
  <action id="com.example.foo.greet">
    <_description>yes9</_description>
    <_message>yes10</_message>
    <defaults>
      <allow_active>no6</allow_active>
    </defaults>
  </action>
</policyconfig>""",
        )

        self._mksrc(
            "gtk/test.ui",
            """<?xml version="1.0"?>
<interface>
  <requires lib="gtk+" version="2.16"/>
  <object class="GtkWindow" id="window1">
    <property name="title" translatable="yes">yes11</property>
    <child><placeholder/></child>
  </object>
</interface>""",
        )

        self._mksrc(
            "data/settings.ui",
            """<?xml version="1.0"?>
<interface domain="foobar">
  <requires lib="gtk+" version="2.16"/>
  <object class="GtkWindow" id="window1">
    <property name="title" translatable="yes">yes12</property>
    <child><placeholder/></child>
  </object>
</interface>""",
        )

        self._mksrc("Makefile", 'echo _("no7")')

        # Executables without *.py extension
        self._mksrc(
            "gtk/foo-gtk", '#!/usr/bin/python\nprint (_("yes13"))', executable=True
        )
        self._mksrc(
            "cli/foo-cli", "#!/usr/bin/env python\nprint (_('yes14'))", executable=True
        )
        self._mksrc("daemon/foobarize", '#!/usr/bin/flex\np _("no8")', executable=True)

    def _src_contents(self, path):
        full_path = pathlib.Path(self.src) / path
        return full_path.read_text("utf-8")

    def _installed_contents(self, path):
        full_path = pathlib.Path(self.install_tree) / path
        return full_path.read_text("utf-8")


if __name__ == "__main__":
    unittest.main()



"""distutils_extra.command.clean_i18n

Implements the Distutils 'clean_i18n' command."""

# TODO: Address following pylint complaints
# pylint: disable=eval-used,missing-function-docstring

import logging
import os
import os.path
import pathlib
import shutil
import typing

from setuptools import Command, Distribution


def _log_error(unused_function: typing.Any, path: str, excinfo: typing.Tuple) -> None:
    logger = logging.getLogger(__name__)
    logger.warning("error removing %s: %s", path, excinfo[1])


# pylint: disable-next=invalid-name
class clean_i18n(Command):
    """clean up files generated by build_i18n"""

    description = __doc__
    user_options = [
        ("build-base=", "b", "base build directory (default: 'build.build-base')"),
        (
            "build-lib=",
            None,
            "build directory for all modules (default: 'build.build-lib')",
        ),
    ]

    def __init__(self, dist: Distribution, **kwargs: dict[str, typing.Any]) -> None:
        super().__init__(dist, **kwargs)
        self.distribution = dist
        self.initialize_options()

    def initialize_options(self) -> None:
        self.build_base: typing.Optional[str] = None

    def finalize_options(self) -> None:
        self.set_undefined_options("build", ("build_base", "build_base"))

    def run(self) -> None:
        """Clean up files generated by build_i18n."""
        assert self.build_base
        # clean build/mo
        mo_dir = os.path.join(self.build_base, "mo")
        if os.path.isdir(mo_dir):
            shutil.rmtree(mo_dir, onerror=_log_error)

        # clean built i18n files
        for setname in (
            "xml_files",
            "desktop_files",
            "schemas_files",
            "rfc822deb_files",
            "ba_files",
            "key_files",
        ):
            file_set = eval(
                self.distribution.get_option_dict("build_i18n").get(
                    setname, (None, "[]")
                )[1]
            )
            for target, files in file_set:
                build_target = os.path.join(self.build_base, target)
                for file in files:
                    if file.endswith(".in"):
                        file_merged = os.path.basename(file[:-3])
                    else:
                        file_merged = os.path.basename(file)
                    file_merged = pathlib.Path(build_target) / file_merged
                    file_merged.unlink(missing_ok=True)
                if os.path.exists(build_target):
                    os.removedirs(build_target)



"""DistUtilsExtra.auto

This provides a setup() method for setuptools and DistUtilsExtra which infers as
many setup() arguments as possible. The idea is that your setup.py only needs
to have the metadata and some tweaks for unusual files/paths, in a "convention
over configuration" paradigm.

This currently supports:

 * Python modules (./*.py, only in root directory)
 * Python packages (all directories with __init__.py)
 * Docbook-XML GNOME help files (help/<language>/{*.xml,*.omf,figures})
 * GtkBuilder and Qt4 user interfaces (*.ui) [installed into
   prefix/share/<projectname>/]
 * D-Bus (*.conf and *.service)
 * GSettings schemas (*.gschema.xml)
 * polkit (*.policy.in)
 * Desktop files (*.desktop.in) [into prefix/share/applications, or
   prefix/share/autostart if they have "autostart" anywhere in the path]
 * KDE4 notifications (*.notifyrc.in)
 * Apport hooks (apport/*) [installed into /usr/share/apport/package-hooks]
 * scripts (all in bin/, and ./<projectname>
 * Auxiliary data files (in data/*) [into prefix/share/<projectname>/]
 * automatic po/POTFILES.in (with all source files which contain _())
 * automatic MANIFEST (everything except swap and backup files, *.pyc, and
   revision control)
 * manpages (*.[0-9])
 * icons (data/icons/<size>/<category>/*.{svg,png})
 * files which should go into /etc (./etc/*, copied verbatim)
 * determining "requires" from import statements in source code
 * determining "provides" from shipped packages and modules

If you follow above conventions, then you don't need any po/POTFILES.in,
./setup.cfg, or ./MANIFEST.in, and just need the project metadata (name,
author, license, etc.) in ./setup.py.
"""

# (c) 2009 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>

# TODO: Address following pylint complaints
# pylint: disable=attribute-defined-outside-init,consider-using-with,eval-used
# pylint: disable=global-variable-not-assigned,invalid-name,missing-docstring
# pylint: disable=redefined-outer-name,too-many-branches,unspecified-encoding

import ast
import fnmatch
import locale
import logging
import os
import pathlib
import shutil
import stat
import sys
import typing
from functools import reduce

import setuptools
import setuptools.command.install
import setuptools.command.sdist
from setuptools import Command, Distribution
from setuptools.errors import FileError

from DistUtilsExtra import __version__ as __pkgversion
from DistUtilsExtra.command import (
    build_extra,
    build_help,
    build_i18n,
    build_icons,
    pylint,
)

__version__ = __pkgversion

# FIXME: global variable, to share with build_i18n_auto
# pylint: disable=global-statement
src = {}
src_all = {}


def setup(**attrs):
    """Auto-inferring extension of standard setuptools.setup()"""
    global src
    global src_all
    src_all = src_find(attrs)
    src = src_all.copy()

    # src_find() removes explicit scripts, but we need them for automatic
    # POTFILE.in building and requires
    src_all.update(set(attrs.get("scripts", [])))

    src_mark(src, "setup.py")
    src_markglob(src, "setup.cfg")

    # mark files in etc/*, handled by install_auto
    # don't install DistUtilsExtra if bundled with a source tarball
    # ignore packaging
    ignore_dirs = ["etc", "DistUtilsExtra", "debian"]

    def should_be_ignored(path: str) -> bool:
        for ignore_dir in ignore_dirs:
            if path.startswith(ignore_dir + os.path.sep):
                return True
        # Also remove files from the .egg-info directory
        return ".egg-info/" in path

    src = {f for f in src if not should_be_ignored(f)}

    __cmdclass(attrs)
    __modules(attrs, src)
    __packages(attrs, src)
    __provides(attrs, src)
    __dbus(attrs, src)
    __gschema(attrs, src)
    __apport_hooks(attrs, src)
    __data(attrs, src)
    __scripts(attrs, src)
    __stdfiles(attrs, src)
    __ui(attrs, src)
    __manpages(attrs, src)

    if "clean" not in sys.argv:
        __requires(attrs, src_all)

    setuptools.setup(**attrs)

    if src:
        print("WARNING: the following files are not recognized by DistUtilsExtra.auto:")
        enc = locale.getpreferredencoding()
        for f in sorted(src):
            # ensure that we can always print the file name
            f_loc = f.encode(enc, errors="replace").decode(enc, errors="replace")
            print(f"  {f_loc}")


#
# parts of setup()
#


def _log_error(unused_function: typing.Any, path: str, excinfo: typing.Tuple) -> None:
    logger = logging.getLogger(__name__)
    logger.warning("error removing %s: %s", path, excinfo[1])


class clean_build_tree(Command):
    description = "clean up build/ directory"
    user_options = [
        ("build-base=", "b", "base build directory (default: 'build.build-base')"),
        (
            "build-lib=",
            None,
            "build directory for all modules (default: 'build.build-lib')",
        ),
        ("all", "a", "remove all build output, not just temporary by-products"),
    ]
    boolean_options = ["all"]

    def __init__(self, dist: Distribution, **kwargs: dict[str, typing.Any]) -> None:
        super().__init__(dist, **kwargs)
        self.initialize_options()

    def initialize_options(self) -> None:
        self.all = False
        self.build_base: typing.Optional[str] = None

    def finalize_options(self) -> None:
        self.set_undefined_options("build", ("build_base", "build_base"))

    def run(self) -> None:
        assert self.build_base
        # clean build/mo
        if os.path.exists(self.build_base):
            logging.getLogger(__name__).info("removing '%s'", self.build_base)
            shutil.rmtree(self.build_base, onerror=_log_error)


def __cmdclass(attrs):
    """Default cmdclass for DistUtilsExtra"""

    v = attrs.setdefault("cmdclass", {})
    v.setdefault("build", build_extra.build_extra)
    v.setdefault("build_help", build_help_auto)
    v.setdefault("build_i18n", build_i18n_auto)
    v.setdefault("build_icons", build_icons.build_icons)
    v.setdefault("install", install_auto)
    v.setdefault("clean", clean_build_tree)
    v.setdefault("sdist", sdist_auto)
    v.setdefault("pylint", pylint.pylint)


def __modules(attrs, src):
    """Default modules"""

    if "py_modules" in attrs:
        for mod in attrs["py_modules"]:
            print(mod)
            src_markglob(src, os.path.join(mod, "*.py"))
        return

    mods = attrs.setdefault("py_modules", [])

    for f in src_fileglob(src, "*.py"):
        if os.path.sep not in f:
            mods.append(os.path.splitext(f)[0])
            src_markglob(src, f)


def __packages(attrs, src):
    """Default packages"""

    if "packages" in attrs:
        for pkg in attrs["packages"]:
            src_markglob(src, os.path.join(pkg, "*.py"))
        return

    packages = attrs.setdefault("packages", [])

    for f in src_fileglob(src, "__init__.py"):
        if f.startswith(f"data{os.path.sep}"):
            continue
        pkg = os.path.dirname(f)
        packages.append(pkg)
        src_markglob(src, os.path.join(pkg, "*.py"))


def __dbus(attrs, src):
    """D-Bus configuration and services"""

    v = attrs.setdefault("data_files", [])

    # /etc/dbus-1/system.d/*.conf
    dbus_conf = []
    for f in src_fileglob(src, "*.conf"):
        if "-//freedesktop//DTD D-BUS Bus Configuration" in open(f).read():
            src_mark(src, f)
            dbus_conf.append(f)
    if dbus_conf:
        v.append(("/etc/dbus-1/system.d/", dbus_conf))

    session_service = []
    system_service = []
    # dbus services
    for f in src_fileglob(src, "*.service"):
        lines = [line.strip() for line in open(f).readlines()]
        if "[D-BUS Service]" not in lines:
            continue
        for line in lines:
            if line.startswith("User="):
                src_mark(src, f)
                system_service.append(f)
                break
        else:
            src_mark(src, f)
            session_service.append(f)
    if system_service:
        v.append(("share/dbus-1/system-services", system_service))
    if session_service:
        v.append(("share/dbus-1/services", session_service))


def __gschema(attrs, src):
    """Install GSettings schema files"""

    v = attrs.setdefault("data_files", [])
    schema_glob = "*.gschema.xml"
    schemas = src_fileglob(src, schema_glob)
    if schemas:
        src_markglob(src, schema_glob)
        src_markglob(src, "*gschemas.compiled")
        v.append(("share/glib-2.0/schemas/", schemas))


def __apport_hooks(attrs, src):
    """Apport hooks"""
    v = attrs.setdefault("data_files", [])

    # files will be copied to /usr/share/apport/package-hooks/
    hooks = []
    assert "name" in attrs, 'You need to set the "name" property in setup.py'
    for f in src_fileglob(src, "*.py"):
        if f.startswith("apport/"):
            hooks.append(f)
            src_mark(src, f)
    if hooks:
        v.append(("share/apport/package-hooks/", hooks))


def __data(attrs, src):
    """Install auxiliary data files.

    This installs everything from data/ except data/icons/ and *.in files (which
    are handled differently) into prefix/share/<projectname>/.
    """
    v = attrs.setdefault("data_files", [])

    assert "name" in attrs, 'You need to set the "name" property in setup.py'

    for f in src.copy():
        if (
            f.startswith("data/")
            and not f.startswith("data/icons/")
            and not f.endswith(".desktop.in")
            and not f.endswith(".notifyrc.in")
        ):
            if not os.path.islink(f):
                # symlinks are handled in install_auto
                v.append(
                    (os.path.join("share", attrs["name"], os.path.dirname(f[5:])), [f])
                )
            src_mark(src, f)


def __scripts(attrs, src):
    """Install scripts.

    This picks executable scripts in bin/*, and an executable ./<projectname>.
    Other scripts have to be added manually; this is to avoid automatically
    installing test suites, build scripts, etc.
    """
    assert "name" in attrs, 'You need to set the "name" property in setup.py'

    scripts = []
    for f in src.copy():
        if f.startswith("bin/") or f == attrs["name"]:
            st = os.lstat(f)
            if stat.S_ISREG(st.st_mode) and st.st_mode & stat.S_IEXEC:
                scripts.append(f)
                src_mark(src, f)
            elif stat.S_ISLNK(st.st_mode):
                # symlinks are handled in install_auto
                src_mark(src, f)

    if scripts:
        v = attrs.setdefault("scripts", [])
        v += scripts


def __stdfiles(attrs, src):
    """Install/mark standard files.

    This covers COPYING, AUTHORS, README, etc.
    """
    src_markglob(src, "COPYING*")
    src_markglob(src, "LICENSE*")
    src_markglob(src, "AUTHORS")
    src_markglob(src, "MANIFEST.in")
    src_markglob(src, "MANIFEST")
    src_markglob(src, "TODO")

    # install all README* from the root directory
    readme = []
    for f in src_fileglob(src, "README*").union(src_fileglob(src, "NEWS")):
        if os.path.sep not in f:
            readme.append(f)
            src_mark(src, f)
    if readme:
        assert "name" in attrs, 'You need to set the "name" property in setup.py'

        attrs.setdefault("data_files", []).append(
            (os.path.join("share", "doc", attrs["name"]), readme)
        )


def __ui(attrs, src):
    """Install GtkBuilder/Qt *.ui files"""

    ui = []
    for f in src_fileglob(src, "*.ui"):
        fd = open(f, "rb")
        firstlines = fd.readline()
        firstlines += b"\n" + fd.readline()
        firstlines += b"\n" + fd.readline()
        fd.close()
        if b"<interface" in firstlines or b"<ui version=" in firstlines:
            src_mark(src, f)
            ui.append(f)
    if ui:
        assert "name" in attrs, 'You need to set the "name" property in setup.py'

        attrs.setdefault("data_files", []).append(
            (os.path.join("share", attrs["name"]), ui)
        )


def __manpages(attrs, src):
    """Install manpages"""

    mans = {}
    for f in src_fileglob(src, "*.[0123456789]"):
        with open(f) as fd:
            for line in fd:
                if line.startswith('."'):
                    continue
                if line.startswith(".TH "):
                    src_mark(src, f)
                    mans.setdefault(f[-1], []).append(f)
                break

    v = attrs.setdefault("data_files", [])
    for section, files in mans.items():
        v.append((os.path.join("share", "man", f"man{section}"), files))


# pylint: disable-next=too-many-return-statements
def __external_mod(cur_module, module, attrs):
    """Check if given Python module is not included in Python or locally"""

    # filter out locally provided modules early, to avoid importing them (which
    # might raise an exception, or parse argv, etc.)
    if module in attrs["provides"]:
        return False
    for m in _module_parents(module):
        if m in attrs["provides"]:
            return False

    try:
        mod = __import__(module)
    except ImportError:
        # try relative import
        try:
            if cur_module:
                mod = __import__(f"{cur_module}.{module}")
            else:
                raise
        except ImportError:
            sys.stderr.write(f"ERROR: Python module {module} not found\n")
            return False
        except (RuntimeError, ValueError):
            # RuntimeError: When Gdk can't be initialized
            # ValueError: weird ctypes case with wintypes
            return False
    except (RuntimeError, ValueError):
        # RuntimeError: When Gdk can't be initialized
        # ValueError: weird ctypes case with wintypes
        return False

    if not hasattr(mod, "__file__") or not mod.__file__:
        # builtin module
        return False

    # filter out locally provided modules
    if mod.__name__ in attrs["provides"]:
        return False

    return (
        "dist-packages" in mod.__file__
        or "site-packages" in mod.__file__
        or not mod.__file__.startswith(os.path.dirname(os.__file__))
    )


def __add_imports(imports, file, attrs):
    """Add all imported modules from file to imports set.

    This filters out modules which are shipped with Python itself.
    """
    if os.path.exists(os.path.join(os.path.dirname(file), "__init__.py")):
        cur_module = ".".join(file.split(os.path.sep)[:-1])
    else:
        # this might happen for paths like bin/<script> which we do not want to
        # treat as module for checking relative imports
        cur_module = None

    try:
        with open(file, "rb") as f:
            file_content = f.read().decode("UTF-8")
            tree = ast.parse(file_content, file)

        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    if alias.name and __external_mod(cur_module, alias.name, attrs):
                        imports.add(alias.name)
            if isinstance(node, ast.ImportFrom):
                if node.module == "gi.repository":
                    for name in node.names:
                        imports.add(f"gi.repository.{name.name}")

                elif node.module and __external_mod(cur_module, node.module, attrs):
                    imports.add(node.module)
    except SyntaxError as e:
        sys.stderr.write(f"WARNING: syntax errors in {file}: {str(e)}\n")


def _module_parents(mod):
    """Iterate over all parents of a module"""

    hierarchy = mod.split(".")
    hierarchy.pop()
    while hierarchy:
        yield ".".join(hierarchy)
        hierarchy.pop()


def __filter_namespace(modules):
    """Filter out modules which are already covered by a parent module

    E. g. this transforms ['os.path', 'os', 'foo.bar.baz', 'foo.bar'] to
    ['os', 'foo.bar'].
    """
    result = set()

    for m in modules:
        if m.startswith("gi.repository."):
            result.add(m)
            continue
        for p in _module_parents(m):
            if p in modules:
                break
        else:
            result.add(m)

    return sorted(result)


def __requires(attrs, src_all):
    """Determine requires (if not set explicitly)"""

    if "requires" in attrs:
        return

    imports = set()

    # iterate over all *.py and scripts which are Python
    for s in src_all:
        if s == "setup.py":
            continue
        if s.startswith(f"data{os.path.sep}"):
            continue
        ext = os.path.splitext(s)[1]
        if ext == "":
            try:
                with open(s) as f:
                    line = f.readline()
            except (UnicodeDecodeError, IOError):
                continue
            if not line.startswith("#!") or "python" not in line:
                continue
        elif ext != ".py":
            continue
        __add_imports(imports, s, attrs)

    attrs["requires"] = __filter_namespace(imports)


def __provides(attrs, unused_src_all):
    """Determine provides (if not set explicitly)"""

    if "provides" in attrs:
        return

    provides = list(attrs.get("py_modules", []))  # we need a copy here
    for p in attrs.get("packages", []):
        provides.append(p.replace(os.path.sep, "."))
    attrs["provides"] = __filter_namespace(provides)


#
# helper functions
#


def src_find(attrs):
    """Find source files.

    This ignores all source files which are explicitly specified as setup()
    arguments.
    """
    src = set()

    # files explicitly covered in setup() call
    explicit = set(attrs.get("scripts", []))
    for _, files in attrs.get("data_files", []):
        explicit.update(files)

    for root, _, files in os.walk("."):
        if root.startswith("./"):
            root = root[2:]
        if root == ".":
            root = ""
        if root.startswith(".") or root.split(os.path.sep, 1)[0] in (
            "build",
            "test",
            "tests",
        ):
            continue
        # data/icons is handled by build_icons
        if root.startswith(os.path.join("data", "icons")):
            continue
        for f in files:
            ext = os.path.splitext(f)[1]
            if f.startswith(".") or ext in (".pyc", "~", ".mo"):
                continue
            # po/*.po is taken care of by build_i18n
            if root == "po" and (ext == ".po" or f == "POTFILES.in"):
                continue

            path = os.path.join(root, f)
            if path not in explicit:
                src.add(path)

    return src


def src_fileglob(src, fnameglob):
    """Return set of files which match fnameglob."""

    result = set()
    for f in src:
        if fnmatch.fnmatch(os.path.basename(f), fnameglob):
            result.add(f)
    return result


def src_mark(src, path):
    """Remove path from src."""

    src.remove(path)


def src_markglob(src, pathglob):
    """Remove all paths from src which match pathglob."""

    for f in src.copy():
        if fnmatch.fnmatch(f, pathglob):
            src.remove(f)


#
# Automatic setup.cfg
#


class build_help_auto(build_help.build_help):
    def finalize_options(self):
        super().finalize_options()

        for data_set in self.get_data_files():
            for filepath in data_set[1]:
                src.remove(filepath)


class build_i18n_auto(build_i18n.build_i18n):
    def finalize_options(self):
        super().finalize_options()

        # add polkit files
        policy_files = []
        for f in src_fileglob(src, "*.policy.in"):
            src_mark(src, f)
            policy_files.append(f)
        if policy_files:
            try:
                xf = eval(self.xml_files)
            except TypeError:
                xf = []
            xf.append((os.path.join("share", "polkit-1", "actions"), policy_files))
            self.xml_files = repr(xf)

        # add desktop files
        desktop_files = []
        autostart_files = []
        notify_files = []
        for f in src_fileglob(src, "*.desktop.in"):
            src_mark(src, f)
            if "autostart" in f:
                autostart_files.append(f)
            else:
                desktop_files.append(f)
        for f in src_fileglob(src, "*.notifyrc.in"):
            src_mark(src, f)
            notify_files.append(f)
        try:
            df = eval(self.desktop_files)
        except TypeError:
            df = []
        if desktop_files:
            df.append(("share/applications", desktop_files))
        if autostart_files:
            df.append(("share/autostart", autostart_files))
        if notify_files:
            df.append((f"share/kde4/apps/{self.distribution.get_name()}", notify_files))
        self.desktop_files = repr(df)

        # mark PO template as known to handle
        try:
            src_mark(
                src, os.path.join(self.po_dir, f"{self.distribution.get_name()}.pot")
            )
        except KeyError:
            pass

    def run(self):
        """Build a default POTFILES.in"""

        auto_potfiles_in = False
        exe_symlinks = []
        try:
            if not os.path.exists(os.path.join("po", "POTFILES.in")):
                prefix = {}
                files = src_fileglob(src_all, "*.py")
                files.update(src_fileglob(src_all, "*.desktop.in"))
                files.update(src_fileglob(src_all, "*.notifyrc.in"))
                files.update(src_fileglob(src_all, "*.policy.in"))

                for f in src_fileglob(src_all, "*.ui"):
                    contents = open(f, "rb").read()
                    if (
                        b"<interface>\n" in contents or b"<interface " in contents
                    ) and b'class="Gtk' in contents:
                        prefix[f] = "[type: gettext/glade]"
                        files.add(f)

                # find extensionless executable scripts which are Python files, and
                # generate a temporary *.py alias, so that they get caught by
                # intltool
                for f in reduce(
                    lambda x, y: x.union(y[1]), self.distribution.data_files, src_all
                ):
                    f_py = f"{f}.py"
                    if (
                        os.access(f, os.X_OK)
                        and os.path.splitext(f)[1] == ""
                        and not os.path.exists(f_py)
                    ):
                        line = open(f, "rb").readline()
                        if line.startswith(b"#!") and b"python" in line:
                            os.symlink(os.path.basename(f), f_py)
                            files.add(f_py)
                            exe_symlinks.append(f_py)

                if files:
                    if not os.path.isdir("po"):
                        os.mkdir("po")
                    potfiles_in = open("po/POTFILES.in", "w")
                    potfiles_in.write("[encoding: UTF-8]\n")
                    for f in sorted(files):
                        potfiles_in.write(f"{prefix.get(f, '') + f}\n")
                    potfiles_in.close()

                    auto_potfiles_in = True

            super().run()
        finally:
            for f in exe_symlinks:
                os.unlink(f)

        if auto_potfiles_in:
            os.unlink("po/POTFILES.in")
            try:
                os.rmdir("po")
            except OSError:
                pass


#
# Automatic sdist
#


class sdist_auto(setuptools.command.sdist.sdist):
    """Default values for the 'sdist' command.

    Replace the manually maintained MANIFEST.in file by providing information
    about what the source tarball created using the 'sdist' command should
    contain in normal cases.

    It prevents the 'build' directory, version control related files, as well as
    compiled Python and gettext files and temporary files from being included in
    the source tarball.

    It's possible for subclasses to extend the 'filter_prefix' and
    'filter_suffix' properties.
    """

    filter_prefix = ["build", ".git", ".svn", ".CVS", ".bzr", ".shelf"]
    filter_suffix = [".pyc", ".mo", "~", ".swp"]

    def __init__(self, dist: Distribution, **kwargs: dict[str, typing.Any]) -> None:
        super().__init__(dist, **kwargs)
        self.distribution = dist
        self.sub_commands.append(("add_extra", sdist_auto.add_extra))

    def add_extra(self) -> None:
        if os.path.exists("MANIFEST.in"):
            return

        self.filter_prefix.append(os.path.join("dist", self.distribution.get_name()))

        for path in pathlib.Path(".").rglob("*"):
            f = str(path)
            if (
                path.is_dir()
                or f in self.filelist.files
                or any(map(f.startswith, self.filter_prefix))
                or any(map(f.endswith, self.filter_suffix))
            ):
                continue

            self.filelist.append(f)


#
# Automatic installation of ./etc/ and symlinks
#


def copy_tree(src: pathlib.Path, dst: pathlib.Path) -> None:
    """Copy an entire directory tree 'src' to a new location 'dst'.

    Both 'src' and 'dst' must be directory names.  If 'src' is not a
    directory, raise FileError.  If 'dst' does not exist, it is
    created.  The end result of the copy is that every file in 'src'
    is copied to 'dst', and directories under 'src' are recursively
    copied to 'dst'.
    """
    log = logging.getLogger(__name__)

    if not src.is_dir():
        raise FileError(f"cannot copy tree '{src}': not a directory")

    if not dst.exists():
        log.info("creating %s", dst)
        dst.mkdir(parents=True, exist_ok=True)

    for src_path in sorted(src.iterdir()):
        dst_path = dst / src_path.name

        if src_path.is_dir():
            copy_tree(src_path, dst_path)

        elif src_path.is_symlink():
            link_dest = src_path.readlink()
            log.info("linking %s -> %s", dst_path, link_dest)
            try:
                dst_path.unlink(missing_ok=True)
            except OSError as error:
                raise FileError(f"could not delete '{dst_path}': {error}") from error
            dst_path.symlink_to(link_dest)

        else:
            log.info("copying %s -> %s", src_path, dst_path)
            try:
                dst_path.unlink(missing_ok=True)
            except OSError as error:
                raise FileError(f"could not delete '{dst_path}': {error}") from error
            shutil.copy(src_path, dst_path)


class install_auto(setuptools.command.install.install):
    def run(self):
        # run build_* subcommands to get file lists if install command
        # won't run 'build' for us
        if self.skip_build:
            self.run_command("build_help")
            self.run_command("build_i18n")
            self.run_command("build_icons")
        self._install_files_from_etc()
        self._install_data_and_scripts_symlinks()
        super().run()

    def _install_files_from_etc(self) -> None:
        if os.path.isdir("etc"):
            # work around a bug in copy_tree() which fails with "File exists" on
            # previously existing symlinks
            for path in pathlib.Path("etc").rglob("*"):
                if path.is_dir():
                    continue
                f = str(path)
                if not f.startswith(f"etc{os.path.sep}") or not os.path.islink(f):
                    continue
                try:
                    os.unlink(os.path.join(self.root, f))
                except OSError:
                    pass
            if not self.root:
                self.root = ""
            copy_tree(pathlib.Path("etc"), pathlib.Path(os.path.join(self.root, "etc")))

    def _install_data_and_scripts_symlinks(self) -> None:
        for path, _, files in os.walk("."):
            for f in files:
                f = os.path.join(path, f)
                if not os.path.islink(f):
                    continue

                if f.startswith("./bin/") or f.startswith("./data/"):
                    if f.startswith("./bin"):
                        target_dir = self.install_scripts
                        dest = os.path.join(
                            target_dir, os.path.sep.join(f.split(os.path.sep)[2:])
                        )
                    elif f.startswith("./data/icons"):
                        target_dir = os.path.join(
                            self.install_data, "share", "icons", "hicolor"
                        )
                        dest = os.path.join(
                            target_dir, os.path.sep.join(f.split(os.path.sep)[3:])
                        )
                    else:
                        target_dir = os.path.join(
                            self.install_data, "share", self.distribution.get_name()
                        )
                        dest = os.path.join(
                            target_dir, os.path.sep.join(f.split(os.path.sep)[2:])
                        )

                    d = os.path.dirname(dest)
                    if not os.path.isdir(d):
                        os.makedirs(d)
                    pathlib.Path(dest).unlink(missing_ok=True)
                    os.symlink(os.readlink(f), dest)
